from pathlib import Path
from time import sleep
import pytest
from scsm.core import App, Server, SteamCMD

APP_ID = 232370
APP_NAME = 'hl2dm'
APP_ID_FAIL = 380840
APP_NAME_FAIL = 'teeworld'

APP_DIR = Path('~/.local/share/scsm/apps').expanduser()
BACKUP_DIR = Path('~/.local/share/scsm/backups').expanduser()


@pytest.fixture
def app():
    app = App(APP_ID, APP_DIR, BACKUP_DIR)
    return app


@pytest.fixture
def app_installed(app, steamcmd_installed):
    if not app.installed:
        app.update()
    return app


@pytest.fixture
def app_removed(app, server_stopped):
    if app.installed:
        app.remove()
    return app


@pytest.fixture
def server():
    server = Server(APP_NAME, APP_DIR, BACKUP_DIR)
    return server


@pytest.fixture
def server_running(server, app_installed):
    if not server.running:
        server.start()
        sleep(5)
    return server


@pytest.fixture
def server_stopped(server, app_installed):
    if server.running:
        server.stop()
        for _ in range(10):
            if not server.running:
                break
            sleep(1)
        else:
            server.kill()
    return server


@pytest.fixture
def steamcmd():
    steamcmd = SteamCMD()
    return steamcmd


@pytest.fixture
def steamcmd_installed(steamcmd):
    if not steamcmd.installed:
        steamcmd.install()
        steamcmd.update()
    return steamcmd


@pytest.fixture
def steamcmd_removed(steamcmd):
    if steamcmd.installed and steamcmd.exe != 'steamcmd':
        steamcmd.remove()
    return steamcmd



import os
import platform as pf
import re
import shutil
import subprocess
import tarfile
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile
from urllib.request import urlretrieve

import vdf
from ruamel.yaml import YAML

from .config import Config


class App():
    def __init__(self, app, app_dir, backup_dir=None, platform=None):
        self.app_id, self.app_name, self.server_name = Index.search(app)

        # app_id is required, if None raise FileNotFoundError
        f = f'{self.app_id}.yaml'
        self.config_f = Path(Config.data_dir, 'apps', f)
        self.config_is_default = True

        # check for user app config file
        d = Path(Config.config_dir, 'apps')
        if d.exists() and Path(d, f).is_file():
            self.config_f = Path(d, f)
            self.config_is_default = False

        with open(self.config_f, 'r') as f:
            yaml = YAML(typ='safe')
            data = yaml.load(f)

        self.app_names = list(data['apps'].keys())
        if not self.app_name:
            self.app_name = self.app_names[0]

        self.server_names = list(data['apps'][self.app_name]['servers'].keys())

        data = data['apps'][self.app_name]
        self.full_name = data['fname']

        if self.server_name:
            self.start_options = data['servers'][self.server_name]['start']
            self.stop_options = data['servers'][self.server_name]['stop']

        self.app_dir = Path(app_dir, str(self.app_id), self.app_name)
        if backup_dir:
            self.backup_dir = Path(backup_dir, str(self.app_id), self.app_name)

        self.beta, self.beta_password, self.app_config = None, None, None
        try:
            self.beta = data['beta']
            self.beta_password = data['password']
            self.app_config = data['app_config']
        except KeyError:
            pass

        if not platform:
            self.platform = pf.system()
        else:
            self.platform = platform

        self.platforms = data['platforms'].keys()
        self.arch = pf.architecture()[0]

        if self.platform in self.platforms:
            if 'exec' in data['platforms'][self.platform].keys():
                data = data['platforms'][self.platform]
            else:
                data = data['platforms'][self.platform][self.arch]

            self.exe = data['exec']

            if 'directory' in data.keys():
                self.exec_dir = Path(self.app_dir, data['directory'])
            else:
                self.exec_dir = self.app_dir

    @property
    def build_id_local(self):
        '''Return the app's local build id'''
        f = Path(self.app_dir, 'steamapps', f'appmanifest_{self.app_id}.acf')

        if f.is_file():
            with open(f, 'r') as app_manifest:
                data = vdf.load(app_manifest)
            return int(data['AppState']['buildid'])
        return 0

    @property
    def build_id_steam(self):
        '''Return the app's steam build id'''
        steamcmd = SteamCMD()
        data = steamcmd.info(self.app_id)

        if data:
            if self.beta:
                return int(data['depots']['branches'][self.beta]['buildid'])
            return int(data['depots']['branches']['public']['buildid'])
        return 0

    @property
    def installed(self):
        '''Return True if app is installed'''
        if self.app_dir.exists():
            for d in self.app_dir.iterdir():
                # if only steamapps directory is found
                # then it did not completely install
                if d != 'steamapps':
                    return True
        return False

    @property
    def running(self):
        '''Return True if app is running'''
        # Server functions do not work on Windows
        if self.platform != 'Windows':
            return Server.running_check(self.app_name)
        return False

    def backup(self, compression=None):
        '''Backup app to backup_dir using tar'''
        if not compression:
            compression = ''
            extension = '.tar'
        else:
            extension = f'.tar.{compression}'

        date = datetime.now().strftime("%Y-%m-%d-%H%M%S")
        f = Path(self.backup_dir, f'{date}{extension}')

        with tarfile.open(f, f'w:{compression}') as tar:
            os.chdir(self.app_dir.parent)
            tar.add(self.app_name)

    def copy_config(self):
        '''Copy default app config file to config_dir'''
        f = f'{self.app_id}.yaml'
        shutil.copyfile(Path(Config.data_dir, 'apps', f),
                        Path(Config.config_dir, 'apps', f))
        self.config_is_default = False

    def remove(self):
        '''Remove app directory'''
        shutil.rmtree(self.app_dir)

        # if this app is the only one installed for that app_id
        # remove the now empty app_id directory as well
        app_dir = self.app_dir.parent
        if not os.listdir(app_dir):
            app_dir.rmdir()

    def restore(self, backup):
        '''Restore specified backup file'''
        with tarfile.open(Path(self.backup_dir, backup)) as tar:
            tar.extractall(self.app_dir.parent)

        if self.config_is_default:
            self.copy_config()

    def update(self, username='anonymous', password='',
               steam_guard='', validate=False, verbose=False):
        '''Update app using steamcmd'''
        new_install = not self.app_dir.exists()

        if self.config_is_default:
            self.copy_config()

        steamcmd = SteamCMD()
        exit_code, text = steamcmd.app_update(self.app_id, self.app_dir,
                                              self.beta, self.beta_password,
                                              self.app_config, self.platform,
                                              validate, username, password,
                                              steam_guard, verbose)

        # remove partially downloaded files from no subscription error
        if '(No subscription)' in text and new_install:
            self.remove()

        return exit_code, text


class Index():
    '''Used for working with app_index.yaml'''
    f = Path(Config.config_dir, 'app_index.yaml')

    @staticmethod
    def config_dirs():
        '''Get app config dirs'''
        for d in Config.data_dir, Config.config_dir:
            d = Path(d, 'apps')
            if d.exists():
                yield d

    @staticmethod
    def list(directory):
        '''Return appid or app_name if not only app for app_id'''
        directory = Path(directory)
        if not directory.exists():
            return

        with open(Index.f, 'r') as f:
            yaml = YAML(typ='safe')
            data = yaml.load(f)

        for app_id in directory.iterdir():
            if len(data[int(app_id.name)].keys()) > 1:
                for app_name in Path(directory, app_id).iterdir():
                    yield app_name.name
            else:
                yield app_id.name

    @staticmethod
    def list_all():
        '''Return generator of all app_id's in index'''
        with open(Index.f, 'r') as f:
            yaml = YAML(typ='safe')
            data = yaml.load(f)

        for app_id in data.keys():
            app_names = data[app_id].keys()

            if len(app_names) > 1:
                for app_name in app_names:
                    yield app_name
            else:
                yield app_id

    @staticmethod
    def search(app):
        '''Search index for app and return app_id, app_name, and server_name'''
        try:
            app = int(app)
        except ValueError:
            pass

        with open(Index.f, 'r') as f:
            yaml = YAML(typ='safe')
            data = yaml.load(f)

        if app in data.keys():
            return app, None, None

        for app_id in data.keys():
            if app in data[app_id].keys():
                if app in data[app_id][app]:
                    return app_id, app, app
                return app_id, app, None
            for app_name in data[app_id].keys():
                if app in data[app_id][app_name]:
                    return app_id, app_name, app
        return None, None, None

    @staticmethod
    def update():
        '''Update index with latst app config files'''
        app_index = {}
        for d in Index.config_dirs():
            for f in d.iterdir():
                if Path(f).suffix == '.yaml':
                    with open(Path(d, f), 'r') as config_f:
                        yaml = YAML(typ='safe')
                        data = yaml.load(config_f)

                    for app in data['apps'].keys():
                        app_index[data['app_id']] = {app: list(data['apps'][app]
                                                               ['servers'].keys())}

        with open(Index.f, 'w') as f:
            yaml = YAML(typ='safe')
            yaml.dump(app_index, f)


class Server(App):
    def __init__(self, app, app_dir, backup_dir=None,  platform=None):
        super(Server, self).__init__(app, app_dir, backup_dir, platform)
        if not self.server_name:
            self.server_name = self.server_names[0]
        self.session = f'{self.app_name}-{self.server_name}'

    @property
    def running(self):
        '''Return True if app is running'''
        if self.server_name == self.app_name:
            return Server.running_check(self.app_name)
        return Server.running_check(self.app_name, self.server_name)

    def console(self):
        '''Attach to screen session'''
        cmd = ['screen', '-r', self.session]
        return subprocess.run(cmd, shell=False).returncode

    def kill(self):
        '''Kill screen session'''
        cmd = ['screen', '-S', self.session, '-X', 'quit']
        subprocess.run(cmd, shell=False)

    @staticmethod
    def running_check(app_name, server_name=None):
        '''Check if server or app is running'''
        if server_name:
            session = f'{app_name}-{server_name}'
        else:
            session = f'{app_name}-.*'

        proc = subprocess.run(['screen', '-ls'], stdout=subprocess.PIPE, shell=False)

        for line in proc.stdout.decode().split('\n'):
            if re.search(fr'{session}\s', line):
                return True
        return False

    def send(self, command):
        '''Send command to screen session'''
        cmd = ['screen', '-S', self.session, '-X', 'stuff', f'{command}\n']
        return subprocess.run(cmd, shell=False).returncode

    def start(self, debug=False):
        '''Start server'''
        if debug:
            cmd = []
        else:
            cmd = ['screen', '-dmS', self.session]

        cmd.extend(self.exe.split(' '))

        # unreal engine games have options that end with a ?
        # that have to be combined into 1 long string
        if self.start_options[0].endswith('?'):
            for i, option in enumerate(self.start_options):
                if i != 0 and self.start_options[i-1].endswith('?'):
                    cmd[-1] += option
                else:
                    cmd.append(option)
        else:
            cmd.extend(self.start_options)

        os.chdir(self.exec_dir)
        subprocess.run(cmd, shell=False)

    def stop(self):
        '''Stop server'''
        if self.stop_options:
            for command in self.stop_options:
                self.send(command)
        else:
            # Send Ctrl - C to screen session
            self.send('$\'\003\'')


class SteamCMD():
    def __init__(self):
        if shutil.which('steamcmd'):
            self.exe = 'steamcmd'
        else:
            if pf.system() != 'Windows':
                self.directory = Path('~/.local/share/scsm/steamcmd').expanduser()
                self.exe = Path(self.directory, 'steamcmd.sh')
            else:
                self.directory = Path(os.getenv('APPDATA'), 'scsm', 'steamcmd')
                self.exe = Path(self.directory, 'self.exe')

    @property
    def installed(self):
        '''Return True if installed'''
        if self.exe == 'steamcmd':
            return True
        return self.directory.exists() and self.exe.is_file()

    def app_update(self, app_id, app_dir, beta=None, beta_password=None,
                   config=None, platform=None, validate=False,
                   username='anonymous', password='', steam_guard='',
                   verbose=False):
        '''+app_update wrapper'''
        cmd = ['+login', username, password, steam_guard, '+force_install_dir',
               app_dir, '+app_update', str(app_id), '+quit']

        if config:
            cmd.insert(-3, f'+app_set_config {app_id} {config}')
        if beta:
            cmd.insert(-1, f'-beta {beta}')
        if beta_password:
            cmd.insert(-1, f'-betapassword {beta_password}')
        if validate:
            cmd.insert(-1, 'validate')

        if platform and platform != pf.system():
            if platform == 'Darwin':
                platform = 'macos'
            elif platform == 'Linux':
                platform = 'linux'
            elif platform == 'Windows':
                platform = 'windows'

            cmd.insert(1, f'+@sSteamCmdForcePlatformType {platform}')

        return self.run(cmd, verbose=verbose)

    def cached_login(self, username):
        '''Check if user has a cached login'''
        cmd = [self.exe, '+login', username, '+quit']
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, stdin=subprocess.DEVNULL,
                              timeout=5, shell=False)

        for line in proc.stdout.decode().split('\n'):
            if 'Using cached credentials' in line:
                return True
        return False

    def filter(self, cmd):
        '''Run command and filter the output'''
        success = ['Success', 'Update complete']
        error = ['ERROR', 'Failed', 'Fatal Error']

        proc = subprocess.run(cmd, stdout=subprocess.PIPE, shell=False)

        for line in reversed(proc.stdout.decode().split('\n')):
            if any(word in line for word in success + error):
                return proc.returncode, line
        return proc.returncode, None

    def info(self, app_id):
        ''''Return app info as dict'''
        cmd = [self.exe, '+login', 'anonymous', '+app_info_update', '1',
               '+app_info_print', str(app_id), '+quit']

        out = subprocess.run(cmd, stdout=subprocess.PIPE, shell=False).stdout.decode().split('\n')
        start, end = 0, 0

        # find the start and end of the vdf file from output
        for i, line in enumerate(out):
            if not start and f'"{app_id}"' == line:
                start = i
                continue
            elif line == '}':
                end = i
                break

        data = '\n'.join(out[start:end + 1])
        return vdf.loads(data)[str(app_id)]

    def install(self):
        '''Install steamcmd'''
        if pf.system() == 'Darwin':
            f = 'steamcmd_osx.tar.gz'
        elif pf.system() == 'Linux':
            f = 'steamcmd_linux.tar.gz'
        elif pf.system() == 'Windows':
            f = 'steamcmd.zip'

        base_url = 'https://steamcdn-a.akamaihd.net/client/installer'
        url = f'{base_url}/{f}'

        self.directory.mkdir(parents=True, exist_ok=True)
        urlretrieve(url, Path(self.directory, f))

        if pf.system() != 'Windows':
            with tarfile.open(Path(self.directory, f)) as tar:
                tar.extractall(self.directory)
        else:
            with ZipFile(Path(self.directory, f)) as zipf:
                zipf.extractall(self.directory)

    def license(self, app_id, username='anonymous', password='', steam_guard=''):
        '''Check if user has a license for app_id'''
        cmd = [self.exe, '+login', username, password, steam_guard,
               '+licenses_for_app', str(app_id), '+quit']

        out = subprocess.run(cmd, stdout=subprocess.PIPE, shell=False).stdout.decode()

        for line in out.split('\n'):
            if 'License packageID' in line:
                return True
        return False

    def remove(self):
        '''Remove steamcmd'''
        shutil.rmtree(self.directory)

    def run(self, args, username='anonymous', password='', steamguard='', verbose=False):
        '''Run steamcmd with args and login'''
        args = [self.exe, username, password, steamguard] + args

        if verbose:
            return subprocess.run(args, shell=False).returncode, None
        return self.filter(args)

    def update(self, verbose=False):
        '''Update steamcmd'''
        return self.run(['+quit'], verbose=verbose)



import os
import pytest
from time import sleep

from scsm.core import Index
from scsm.config import Config


Config.create()
Index.update()


class TestApp():
    def test_init(self, app):
        assert type(app.app_name) is str

    def test_build_id_steam(self, app, steamcmd_installed):
        assert app.build_id_steam > 0

    def test_installed(self, app_removed):
        assert app_removed.installed is False

    def test_running(self, server_stopped):
        assert server_stopped.running is False

    def test_update(self, app_installed):
        exit_code, text = app_installed.update()
        assert exit_code == 0

    def test_build_id_local_success(self, app_installed):
        assert app_installed.build_id_local > 0

    def test_build_id_local_failure(self, app_removed):
        assert app_removed.build_id_local == 0

    @pytest.mark.parametrize('compression', ['tar', 'gz'])
    def test_backup(self, app_installed, compression):
        app_installed.backup_dir.mkdir(parents=True, exist_ok=True)
        backups = os.listdir(app_installed.backup_dir)
        app_installed.backup(compression)
        assert len(os.listdir(app_installed.backup_dir)) > len(backups)

    def test_copy_config(self, app):
        app.copy_config()
        assert app.config_is_default is False

    def test_remove(self, app_installed):
        app_installed.remove()
        assert app_installed.installed is False

    def test_restore(self, app_removed):
        backups = os.listdir(app_removed.backup_dir)
        app_removed.restore(backups[0])
        assert app_removed.installed is True


class TestIndex():
    def test_list(self, app):
        assert list(Index.list(app.app_dir.parent.parent))

    def test_list_all(self):
        assert len(list(Index.list_all())) > 0

    def test_update(self):
        Index.f.unlink()
        Index.update()
        assert Index.f.is_file()

    @pytest.mark.parametrize('app,result', [
        (232370, (232370, None, None)),
        ('232370', (232370, None, None)),
        ('hl2dm', (232370, 'hl2dm', 'hl2dm')),
    ])
    def test_search(self, app, result):
        assert Index.search(app) == result


class TestServer():
    def test_init(self, server):
        assert type(server.server_name) is str

    def test_running(self, server_stopped):
        assert server_stopped.running is False

    def test_running_check(self, server_stopped):
        assert server_stopped.running_check(server_stopped.app_name) is False

    def test_start(self, server_stopped):
        server_stopped.start()
        sleep(5)
        assert server_stopped.running is True

    def test_console(self, server):
        assert True

    def test_send(self, server_running):
        assert server_running.send('test') == 0

    def test_stop(self, server_running):
        server_running.stop()

        for i in range(10):
            if not server_running.running:
                break
            sleep(1)

        assert server_running.running is False

    def test_kill(self, server_running):
        server_running.kill()

        for i in range(10):
            if not server_running.running:
                break
            sleep(1)

        assert server_running.running is False


class TestSteamCMD():
    def test_install(self, steamcmd_removed):
        steamcmd_removed.install()
        assert steamcmd_removed.installed is True

    def test_update(self, steamcmd_installed):
        exit_code, text = steamcmd_installed.update()
        assert exit_code == 0

    def test_is_installed(self, steamcmd_installed):
        assert steamcmd_installed.installed is True

    def test_app_update(self, app, steamcmd_installed):
        exit_code, text = steamcmd_installed.app_update(app.app_id, app.app_dir.parent)
        assert exit_code == 0

    def test_cached_login(self, steamcmd_installed):
        assert steamcmd_installed.cached_login('anonymous') is False

    def test_filter(self, steamcmd_installed):
        exit_code, text = steamcmd_installed.filter(['echo', 'Success'])
        assert exit_code == 0

    def test_info(self, app, steamcmd_installed):
        assert type(steamcmd_installed.info(app.app_id)) is dict

    def test_license_true(self, app, steamcmd_installed):
        assert steamcmd_installed.license(app.app_id) is True

    def test_license_false(self, steamcmd_installed):
        assert steamcmd_installed.license(380840) is False

    def test_run(self, steamcmd_installed):
        exit_code, text = steamcmd_installed.run(['+quit'], verbose=False)
        assert exit_code == 0

    def test_run_verbose(self, steamcmd_installed):
        exit_code, text = steamcmd_installed.run(['+quit'], verbose=True)
        assert exit_code == 0

    def test_remove(self, steamcmd_installed):
        steamcmd_installed.remove()
        assert steamcmd_installed.installed is False



__all__ = (
    '__title__', '__summary__', '__url__', '__version__',
    '__author__', '__email__', '__license__'
)

__title__ = 'scsm'
__summary__ = 'SteamCMD Server Mangaer'
__url__ = 'https://github.com/bubylou/scsm'

__version__ = '0.1.8'

__author__ = 'Nicholas Malcolm'
__email__ = 'bubylou7@gmail.com'

__license__ = 'GPLv3+'



import os
import shutil
import textwrap
import types

import pytest
from click.testing import CliRunner

from scsm import cli
from scsm.config import Config


@pytest.fixture(scope='module')
def runner():
    return CliRunner()


def test_backup(runner, server_stopped):
    if server_stopped.backup_dir.exists():
        shutil.rmtree(server_stopped.backup_dir)

    result = runner.invoke(cli.backup, [str(server_stopped.app_id), '--no-compress'])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ App ID ] - 232370
        [ Status ] - Backup started
        [ Status ] - Backup complete
    ''')


def test_console(runner, server_running):
    result = runner.invoke(cli.console, [server_running.app_name], input='$\'\003\'\n')
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Attaching to session
        [ Status ] - Disconnected from session
    ''')


def test_edit(runner, app_installed):
    result = runner.invoke(cli.edit, [str(app_installed.app_id), '--editor', 'more'])
    assert result.exit_code == 0


def test_install(runner, app_removed):
    result = runner.invoke(cli.install, [str(app_removed.app_id)])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent(f'''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ App ID ] - 232370
        [ Status ] - Installing
        [ Status ] - Success! App \'{app_removed.app_id}\' fully installed.
    ''')


def test_kill(runner, server_running):
    result = runner.invoke(cli.kill, [server_running.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Killing
        [ Status ] - Stopped
    ''')


def test_list(runner, app_installed):
    result = runner.invoke(cli.list, [str(app_installed.app_id)])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ F-Name ] - Half-Life 2: Deathmatch
        [ Name   ] - hl2dm
        [ App ID ] - 232370
        [ Status ] - Installed
    ''')


# def test_monitor():
#     result = runner.invoke(cli.monitor, [app_id], input='$\'\003\'\n')
#     assert result.exit_code == 0


def test_remove(runner, server_stopped):
    result = runner.invoke(cli.remove, [str(server_stopped.app_id)])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ App ID ] - 232370
        [ Status ] - Removing
        [ Status ] - Remove complete
    ''')


def test_restart(runner, server_running):
    result = runner.invoke(cli.restart, [server_running.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Stopping
        [ Status ] - Stopped
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Starting
        [ Status ] - Started
    ''')


def test_restore(runner, server_stopped):
    server_stopped.backup_dir.mkdir(parents=True, exist_ok=True)
    if len(os.listdir(server_stopped.backup_dir)) < 1:
        server_stopped.backup(compression=None)

    result = runner.invoke(cli.restore, [str(server_stopped.app_id)], input='1\n')
    assert result.exit_code == 0
    assert result.output.split('\n')[-2] == textwrap.dedent('''\
        [ Status ] - Restore complete''')


def test_send(runner, server_running):
    result = runner.invoke(cli.send, ['test', server_running.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Command sent
        [ Status ] - Command finished
    ''')


def test_setup(runner):
    if Config.config_f.exists() and not Config.system_wide:
        Config.remove()

    result = runner.invoke(cli.setup, input='N\n')
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
         [ ------ ]
         [ Status ] - Creating config files
         [ Status ] - Configs installed
    ''')


def test_start(runner, server_stopped):
    result = runner.invoke(cli.start, [server_stopped.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
         [ ------ ]
         [ Name   ] - hl2dm
         [ Status ] - Starting
         [ Status ] - Started
    ''')


def test_status(runner, server_running):
    result = runner.invoke(cli.status, [server_running.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Running
    ''')


def test_stop(runner, server_running):
    result = runner.invoke(cli.stop, [server_running.app_name])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ Status ] - Stopping
        [ Status ] - Stopped
    ''')


def test_update(runner, app_installed):
    result = runner.invoke(cli.update, [str(app_installed.app_id)])
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
        [ ------ ]
        [ Name   ] - hl2dm
        [ App ID ] - 232370
        [ Status ] - Checking for updates
        [ Status ] - Already up to date
    ''')


@pytest.mark.parametrize('app,result', [
    (('all',), types.GeneratorType),
    (('running',), types.GeneratorType),
    (('backups',), types.GeneratorType)
])
def test_app_special_names(app, result):
    tmp = cli.app_special_names(app)
    assert type(tmp) is result
    assert isinstance(tmp, result)



import os
import platform
from pathlib import Path
from pkg_resources import resource_filename
from ruamel.yaml import YAML


if platform.system() != 'Windows':
    if os.geteuid() == 0:
        BASE_DIR = Path('/opt/scsm')
    else:
        BASE_DIR = Path('~/.local/share/scsm').expanduser()
else:
    BASE_DIR = Path(os.getenv('APPDATA'), 'scsm')


DEFAULTS = f"""
    general:
        compression: gz
        steam_guard: true
        verbose: false
        max_backups: 5
        wait_time: 30
    directories:
        app_dir: {Path(BASE_DIR, 'apps')}
        backup_dir: {Path(BASE_DIR, 'backups')}
    steam:
        username: anonymous
        password:
    """


class Config():
    system_wide = False
    data_dir = resource_filename(__name__, 'data')

    if platform.system() != 'Windows':
        config_dir = Path('~/.config/scsm').expanduser()
        if not config_dir.exists() and Path('/etc/scsm').exists():
            system_wide = True
            config_dir = Path('/etc/scsm')
    else:
        config_dir = Path(os.getenv('APPDATA'), 'scsm')

    _yaml = YAML(typ='safe')
    config_f = Path(config_dir, 'config.yaml')

    if config_f.exists():
        with open(config_f, 'r') as _f:
            data = _yaml.load(_f)
    else:
        data = _yaml.load(DEFAULTS)

    compression = data['general']['compression']
    steam_guard = data['general']['steam_guard']
    verbose = data['general']['verbose']
    max_backups = data['general']['max_backups']
    wait_time = data['general']['wait_time']
    app_dir = Path(data['directories']['app_dir'])
    backup_dir = Path(data['directories']['backup_dir'])
    username = data['steam']['username']
    password = data['steam']['password']

    @staticmethod
    def create(system_wide=False):
        if platform.system() != 'Windows':
            if system_wide:
                config_dir = Path('/etc/scsm')
            else:
                config_dir = Path('~/.config/scsm').expanduser()
        else:
            config_dir = Path(os.getenv('APPDATA'), 'scsm')

        Path(config_dir, 'apps').mkdir(parents=True, exist_ok=True)
        Config.config_dir = config_dir
        Config.config_f = Path(config_dir, 'config.yaml')

        with open(Path(config_dir, 'config.yaml'), 'w') as f:
            yaml = YAML(typ='safe')
            yaml.dump(yaml.load(DEFAULTS), f)

    @staticmethod
    def remove():
        Config.config_f.unlink()



from setuptools import setup, find_packages
import scsm

with open('README.md', 'r') as f:
    README = f.read()

setup(
    name=scsm.__title__,
    version=scsm.__version__,
    url=scsm.__url__,
    description=scsm.__summary__,
    long_description=README,
    long_description_content_type='text/markdown',
    license=scsm.__license__,
    author=scsm.__author__,
    author_email=scsm.__email__,
    packages=find_packages(exclude=['tests']),
    python_requires='>= 3.6',
    install_requires=[
        'click >= 6.6',
        'ruamel.yaml >= 0.15.75',
        'vdf >= 2.4'
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": ['scsm=scsm.cli:main']
        },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities'
    ],
)



import os
import platform
import shutil
import signal
import sys
from email.mime.text import MIMEText
from pathlib import Path
from smtplib import SMTP
from time import sleep

import click

from .config import Config
from .core import App, Index, Server, SteamCMD


LOGIN_OPTIONS = [
    click.option('-u', '--username', default=Config.username, help='Steam username'),
    click.option('-p', '--password', default=Config.password, help='Steam password'),
    click.option('-g', '--steam-guard', default='', help='Steam guard code'),
    click.option('-v', '--verbose', is_flag=True, default=Config.verbose, help='Verbose mode')
]


def add_options(options):
    def wrapper(func):
        for option in reversed(options):
            func = option(func)
        return func
    return wrapper


@click.group()
@click.version_option()
@click.pass_context
def main(ctx):
    signal.signal(signal.SIGINT, signal_handler)

    if ctx.invoked_subcommand != 'setup':
        if not Config.config_f.exists():
            click.echo('[ ------ ]')
            message('Error', 'No config file found')
            ctx.forward(setup)
        Index.update()

    if ctx.invoked_subcommand in ['console', 'kill', 'monitor', 'restart',
                                  'send', 'start', 'status', 'stop']:
        if platform.system() == 'Windows':
            click.echo('[ ------ ]')
            message('Error', 'Not supported on Windows')
            sys.exit(1)

        if not shutil.which('screen'):
            click.echo('[ ------ ]')
            message('Error', 'GNU Screen is not installed')
            sys.exit(1)


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-c', '--compression', default=Config.compression, help='Compression method')
@click.option('-f', '--force', is_flag=True, help='Run command even if running')
@click.option('-n', '--no-compress', is_flag=True, help='No compression')
def backup(apps, compression, no_compress, force):
    '''Backup app'''

    if compression and compression not in ['bz2', 'gz', 'xz']:
        message('Error', 'Invalid compression method')
    else:
        if no_compress:
            compression = None

        for app in app_special_names(apps):
            a = app_wrapper(app)
            info(a.app_name, a.app_id)

            if not a.installed:
                message('Error', 'App not installed')
            elif not force and a.running:
                message('Error', 'Stop server before backup')
            else:
                a.backup_dir.mkdir(parents=True, exist_ok=True)
                backups = os.listdir(a.backup_dir)
                length = len(backups)

                if Config.max_backups != 0 and length >= Config.max_backups:
                    message('Status', 'Max backups reached')
                    message('Status', 'Removing old backups')

                    for backup in backups[0:length - Config.max_backups + 1]:
                        Path(a.backup_dir, backup).unlink()

                message('Status', 'Backup started')
                a.backup(compression)
                message('Status', 'Backup complete')


@main.command()
@click.argument('apps', nargs=-1)
def console(apps):
    '''Attach to server session'''

    for app in app_special_names(apps):
        s = server_wrapper(app)
        info(s.server_name)

        if not s.installed:
            message('Error', 'App not installed')
        elif not s.running:
            message('Error', 'Stopped')
        else:
            message('Status', 'Attaching to session')
            s.console()
            message('Status', 'Disconnected from session')


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-e', '--editor', help='Text editor')
def edit(apps, editor):
    '''Edit config files'''

    if apps == ('config',):
        click.edit(editor=editor, filename=Config.config_f)
    else:
        for app in app_special_names(apps):
            a = app_wrapper(app)

            if not a.installed:
                info(a.app_name, a.app_id)
                message('Error', 'App not installed')
            elif a.config_is_default:
                a.copy_config()
                a = App(a.app_id, a.app_dir)
                click.edit(editor=editor, filename=a.config_f)
                Index.update()
            else:
                click.edit(editor=editor, filename=a.config_f)
                Index.update()


@main.command()
@click.argument('apps', nargs=-1)
@add_options(LOGIN_OPTIONS)
@click.pass_context
def install(ctx, apps, username, password, steam_guard, verbose):
    '''Install app'''

    if apps == ('steamcmd',):
        steamcmd = SteamCMD()
        if steamcmd.installed:
            message('Error', 'SteamCMD is already installed')

            if steamcmd.exe == 'steamcmd':
                message('Error', 'SteamCMD is not managed by SCSM')
            else:
                text = f'[ {click.style("Status", "green")} ] - Reinstall SteamCMD?'
                if click.confirm(text):
                    steamcmd_install(verbose)
        else:
            steamcmd_install(verbose)
    else:
        ctx.forward(update)


@main.command()
@click.argument('apps', nargs=-1)
def kill(apps):
    '''Kill server'''

    for app in app_special_names(apps, server=True):
        s = server_wrapper(app)
        info(s.server_name)

        if not s.installed:
            message('Error', 'App not installed')
        elif not s.running:
            message('Error', 'Stopped')
        else:
            message('Status', 'Killing')
            s.kill()
            message('Status', 'Stopped')


@main.command()
@click.argument('apps', nargs=-1)
def list(apps):
    '''List installed or all installable apps'''

    arg = None

    if apps == ('all',):
        apps = Index.list_all()
    elif apps == ('backups',):
        arg = 'backups'

    for app in app_special_names(apps):
        a = app_wrapper(app)

        click.echo('[ ------ ]')
        message('F-Name', a.full_name)
        message('Name', a.app_name)
        message('App ID', a.app_id)

        if a.installed:
            message('Status', 'Installed')
        else:
            message('Status', 'Not installed')

        if arg == 'backups':
            message('Status', f'Backups (Max {Config.max_backups})')

            backups = os.listdir(a.backup_dir)
            backups.sort(reverse=True)

            for i, backup in enumerate(backups):
                message(i + 1, backup)


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-e', '--email', is_flag=True, help='Email')
@click.option('-r', '--restart', is_flag=True, help='Restart if stopped')
def monitor(apps, email, restart):
    '''Monitor server status'''

    sessions = {}

    for app in app_special_names(apps, server=True):
        try:
            s = Server(app, Config.app_dir)
        except FileNotFoundError:
            info(app)
            message('Error', 'No server entry')
        else:
            if s.server_name:
                sessions[s.session] = {}
            else:
                for server_name in s.server_names:
                    sessions[f'{s.app_name}-{server_name}'] = {}

    sessions = {session: {'running': False, 'restarts': 0, 'time': 0}
                for session in sessions}

    while sessions:
        for session in sessions:
            app_name, server_name = session.split('-')

            if sessions[session]['running']:
                if not Server.running_check(app_name, server_name):
                    sessions[session]['running'] = False

                    info(server_name)
                    message('Status', 'Stopped')

                    if restart:
                        if sessions[session]['restarts'] == 3:
                            message('Error', 'Restarted 3 times in 30 seconds')
                            message('Error', 'Not restarting')

                            if email:
                                from_addr = 'scsm@localhost'
                                to_addr = f'{os.getlogin()}@localhost'
                                text = f''''Server {session} has been restarted more
                                        than 3 times in 30  seconds and will
                                        not be restarted'''

                                msg = MIMEText(text)
                                msg['Subject'] = 'SCSM - Monitor'
                                msg['From'] = from_addr
                                msg['To'] = to_addr

                                e = SMTP('localhost')
                                e.sendmail(from_addr, to_addr, msg.as_string())
                                e.quit()

                        else:
                            sessions[session]['restarts'] += 1
                            start(server_name)
            else:
                if Server.running_check(app_name, server_name):
                    sessions[session]['running'] = True

                    info(server_name)
                    message('Status', 'Running')

            if sessions[session]['time'] == 30:
                sessions[session]['restarts'] = 0
                sessions[session]['time'] = 0
            elif sessions[session]['restarts'] != 0:
                sessions[session]['time'] += 1

        sleep(1)


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-f', '--force', is_flag=True, help='Run command even if running')
def remove(apps, force):
    '''Remove app'''

    if apps == ('steamcmd',):
        steamcmd = SteamCMD()
        if steamcmd.exe == 'steamcmd':
            message('Error', 'SteamCMD is not managed by SCSM')
        else:
            steamcmd.remove()
            message('Status', 'SteamCMD removed')

    else:
        for app in app_special_names(apps):
            a = app_wrapper(app)
            info(a.app_name, a.app_id)

            if not a.installed:
                message('Error', 'App not installed')
            elif not force and a.running:
                message('Error', 'Stop server before validating')
            else:
                message('Status', 'Removing')
                a.remove()
                message('Status', 'Remove complete')


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-w', '--wait-time', type=int, default=Config.wait_time, help='Wait time')
@click.pass_context
def restart(ctx, apps, wait_time):
    '''Restart server'''

    for app in app_special_names(apps, server=True):
        ctx.invoke(stop, apps=[app], wait_time=wait_time)
        ctx.invoke(start, apps=[app], debug=False, verbose=False)


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-f', '--force', is_flag=True, help='Run command even if running')
@click.option('-l', '--latest', is_flag=True, help='Select the latest backup automatically')
def restore(apps, force, latest):
    '''Restore app from backup'''

    for app in app_special_names(apps):
        a = app_wrapper(app)
        info(a.app_name, a.app_id)

        if not a.backup_dir.exists() or not os.listdir(a.backup_dir):
            message('Error', 'No backups found')
        elif not force and a.running:
            message('Error', 'Stop server before restoring')
        else:
            backups = os.listdir(a.backup_dir)
            backups.sort(reverse=True)
            length = len(backups)

            while length > 1 and not latest:
                message('Status', 'Backups')
                for i, backup in enumerate(backups):
                    message(i + 1, backup)
                answer = int(input(f'[ {click.style("Status", "green")} ] - Choose one: '))

                if answer > length or answer < 1:
                    message('Error', 'Invalid selection')
                    click.echo('[ ------ ]')
                else:
                    backup = backups[answer - 1]
                    break
            else:
                backup = backups[0]

            a.app_dir.mkdir(parents=True, exist_ok=True)
            message('Status', 'Restoring')
            a.restore(backup)
            message('Status', 'Restore complete')


@main.command()
@click.argument('command')
@click.argument('apps', nargs=-1)
def send(apps, command):
    '''Send command to server'''

    for app in app_special_names(apps, server=True):
        s = server_wrapper(app)
        info(s.server_name)

        if not s.installed:
            message('Error', 'App not installed')
        elif not s.running:
            message('Error', 'Stopped')
        else:
            message('Status', 'Command sent')
            s.send(command)
            message('Status', 'Command finished')


@main.command()
@click.option('-s', '--system-wide', is_flag=True, default=False, help='System wide config')
def setup(system_wide):
    '''Setup SteamCMD and config files'''

    click.echo('[ ------ ]')

    if Config.config_f and Config.config_f.exists():
        if not system_wide and Config.system_wide:
            message('Status', 'Creating config files')
            Config.create(system_wide)
            message('Status', 'Configs installed')
        else:
            message('Error', 'Config file already exists')

            if click.confirm(f'[ {click.style("Status", "green")} ] - Overwrite config?'):
                message('Status', 'Creating config files')
                message('Status', 'Configs installed')
                Config.create(system_wide)
    else:
        message('Status', 'Creating config files')
        Config.create(system_wide)
        message('Status', 'Configs installed')

    for d in Config.app_dir, Config.backup_dir:
        d.mkdir(parents=True, exist_ok=True)


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-d', '--debug', is_flag=True, help='Debug mode')
@click.option('-v', '--verbose', is_flag=True, default=Config.verbose, help='Verbose mode')
def start(apps, debug, verbose):
    '''Start server'''

    for app in app_special_names(apps, server=True):
        s = server_wrapper(app)
        info(s.server_name)

        if not s.installed:
            message('Error', 'App not installed')
        elif s.running:
            message('Error', 'Already running')
        else:
            message('Status', 'Starting')
            s.start(debug)
            message('Status', 'Started')

            if verbose and not debug:
                sleep(1)
                s.console()


@main.command()
@click.argument('apps', nargs=-1)
def status(apps):
    '''Status server'''

    for app in app_special_names(apps, server=True):
        s = server_wrapper(app)
        info(s.server_name)
        if not s.installed:
            message('Error', 'App not installed')
        elif s.running:
            message('Status', 'Running')
        else:
            message('Status', 'Stopped')


@main.command()
@click.argument('apps', nargs=-1)
@click.option('-w', '--wait-time', type=int, default=Config.wait_time, help='Wait time')
def stop(apps, wait_time):
    '''Stop server'''

    for app in app_special_names(apps, server=True):
        s = server_wrapper(app)
        info(s.server_name)

        if not s.installed:
            message('Error', 'App not installed')
        elif not s.running:
            message('Error', 'Stopped')
        else:
            message('Status', 'Stopping')
            s.stop()

            for i in range(int(wait_time)):
                if not s.running:
                    break
                sleep(1)
            else:
                message('Error', f'Waited {wait_time} seconds')
                message('Error', 'Killing')
                s.kill()

            message('Status', 'Stopped')


@main.command()
@click.argument('apps', nargs=-1)
@add_options(LOGIN_OPTIONS)
@click.option('-c', '--check-only', is_flag=True, help='Check for an update only')
@click.option('-f', '--force', is_flag=True, help='Run command even if running')
@click.option('-n', '--no-check', is_flag=True, help='Don\'t check for an update')
@click.option('-vv', '--validate', is_flag=True, default=False, help='Validate after update')
def update(apps, username, password, steam_guard,
           check_only, no_check, force, validate, verbose):
    '''Update app'''

    if username != 'anonymous' and not verbose:
        click.echo('[ ------ ]')
        message('Status', 'Checking for cached credentials')

        steamcmd = SteamCMD()
        if not steamcmd.cached_login(username):
            message('Status', 'Cached credentials not available')
            message('Alert', 'Login info can be read in OS process list')
            message('Alert', 'To avoid this use -v or --verbose')

            if not password:
                password = click.prompt('password: ', hide_input=True)
            if not steam_guard:
                steam_guard = click.prompt('steam guard: ',)
        else:
            message('Status', 'Using cached credentials')
            password = steam_guard = ''
    else:
        password = steam_guard = ''

    for app in app_special_names(apps):
        a = app_wrapper(app)
        steamcmd_check()
        info(a.app_name, a.app_id)

        if not force and a.running:
            message('Error', 'Stop server before update')
        elif not a.installed:
            message('Status', 'Installing')
            exit_code, text = a.update(username, password, steam_guard,
                                       validate, verbose)

            if not verbose:
                if exit_code == 0:
                    message('Status', text)
                else:
                    message('Error', text)
        else:
            if not no_check and not validate:
                message('Status', 'Checking for updates')

                if a.build_id_local >= a.build_id_steam:
                    message('Status', 'Already up to date')
                    continue
                else:
                    message('Status', 'Update available')

            if not check_only:
                if validate:
                    message('Status', 'Updating and Validating')
                else:
                    message('Status', 'Updating')

                exit_code, text = a.update(username, password, steam_guard,
                                           validate, verbose)

                if not verbose:
                    if exit_code == 0:
                        message('Status', text)
                    else:
                        message('Error', text)


def app_wrapper(app):
    try:
        a = App(app, Config.app_dir, Config.backup_dir)
    except FileNotFoundError:
        click.echo('[ ------ ]')

        try:
            int(app)
        except ValueError:
            message('Name', app)
            message('Error', 'Invalid app name')
        else:
            message('App ID', app)
            message('Error', 'Invalid app id')
        sys.exit(1)
    else:
        return a


def server_wrapper(app):
    try:
        s = Server(app, Config.app_dir)
    except FileNotFoundError:
        click.echo('[ ------ ]')

        try:
            int(app)
        except ValueError:
            message('Name', app)
            message('Error', 'Invalid app or server name')
        else:
            message('App ID', app)
            message('Error', 'Invalid app id')
        sys.exit(1)
    else:
        return s


def app_special_names(apps, server=False):
    if apps in [('all',), ('installed',)]:
        if server:
            apps = server_expand_names(Index.list(Config.app_dir))
        else:
            apps = Index.list(Config.app_dir)
    elif apps in [('running',), ('stopped',)]:
        status = apps[0]
        apps = []
        for app in Index.list(Config.app_dir):
            s = Server(app, Config.app_dir)

            if s.running:
                for server in s.server_names:
                    if Server.running_check(s.app_name, server):
                        if status == 'running':
                            apps.append(server)
                    else:
                        if status == 'stopped':
                            apps.append(server)
            else:
                if status == 'stopped':
                    apps.append(app)
    elif apps == ('backups',):
        apps = Index.list(Config.backup_dir)
    elif server:
        apps = server_expand_names(apps)

    for app in apps:
        yield app


def server_expand_names(apps):
    for app in apps:
        try:
            s = Server(app, Config.app_dir)
        except FileNotFoundError:
            yield app
        else:
            if s.server_name == s.app_name or app == s.app_id:
                for server in s.server_names:
                    yield server
            else:
                yield app


def info(name, app_id=None):
    click.echo('[ ------ ]')
    message('Name', name)
    if app_id:
        message('App ID', app_id)


def message(title, text):
    if title in ['Name', 'App ID', 'F-Name']:
        color = 'yellow'
    elif title in ['Status', 'Done']:
        color = 'green'
    elif title in ['Error', 'Alert']:
        color = 'red'
    else:
        color = 'white'

    click.echo(f'[ {click.style(str(title).ljust(6), color)} ] - {text}')


def signal_handler(signal, frame):
    click.echo(' ')
    message('Status', 'Quitting')
    sys.exit(0)


def steamcmd_check():
    steamcmd = SteamCMD()
    if not steamcmd.installed:
        click.echo('[ ------ ]')
        message('Error', 'SteamCMD not installed')
        steamcmd_install(verbose=False)

    if not steamcmd.installed:
        sys.exit(1)


def steamcmd_install(verbose):
    click.echo('[ ------ ]')
    message('Status', 'SteamCMD installing')

    steamcmd = SteamCMD()
    steamcmd.install()
    message('Status', 'SteamCMD installed')

    message('Status', 'SteamCMD updating')
    exit_code, text = steamcmd.update(verbose)

    if not verbose:
        if exit_code == 0:
            message('Status', 'SteamCMD updated')
        else:
            message('Error', text)
